"""An example of using the LLM Transformer to chain multiple LLM-based transformations together."""

import asyncio

from sklearn.pipeline import make_pipeline

from .base import BaseModel, InputModel
from .transformation import LLMTransformer


class FirstLastName(InputModel):
    first_name: str
    last_name: str


class FullName(InputModel):
    full_name: str


class LowercaseFullName(BaseModel):
    full_name: str


system_prompt1 = "Return the user name by concatenating the first name and last name."
user_prompt1 = """
    First Name: {{first_name}}
    Last Name: {{last_name}}
"""

system_prompt2 = "Return the full name in lowercase."
user_prompt2 = """
    Full Name: {{full_name}}
"""


async def run_separate_transformers(first_last_names: list[FirstLastName]) -> list[LowercaseFullName]:
    """Manually run the two transformers sequentially by feeding output of the first transformer to the second."""
    t1 = LLMTransformer(
        input_type=FirstLastName, output_type=FullName, system_prompt=system_prompt1, user_prompt_template=user_prompt1
    )
    full_names = await t1.transform_async(first_last_names)
    print(full_names)

    t2 = LLMTransformer(
        input_type=FullName,
        output_type=LowercaseFullName,
        system_prompt=system_prompt2,
        user_prompt_template=user_prompt2,
    )
    lowercase_full_names = await t2.transform_async(full_names)
    print(lowercase_full_names)

    return lowercase_full_names


def run_pipeline(first_last_names: list[FirstLastName]) -> list[LowercaseFullName]:
    """Run the two transformers using sklearn's pipeline."""
    pipeline = make_pipeline(
        LLMTransformer(
            input_type=FirstLastName,
            output_type=FullName,
            system_prompt=system_prompt1,
            user_prompt_template=user_prompt1,
        ),
        LLMTransformer(
            input_type=FullName,
            output_type=LowercaseFullName,
            system_prompt=system_prompt2,
            user_prompt_template=user_prompt2,
        ),
    )
    # calling fit_transform will stop the pipeline from complaining about not `fit()` not being called
    lowercase_full_names = pipeline.fit_transform(first_last_names)

    return lowercase_full_names


if __name__ == "__main__":
    first_last_names = [
        FirstLastName(first_name="Trung", last_name="Nguyen"),
        FirstLastName(first_name="John", last_name="Doe"),
    ]
    result1 = run_pipeline(first_last_names)
    result2 = asyncio.run(run_separate_transformers(first_last_names))

    assert result1 == result2



import asyncio
import logging
from typing import Generic, Type

from jinja2 import Template
from sklearn.base import TransformerMixin

from .base import InputType, OutputType
from .llms import OpenAIClient

logging.basicConfig(level=logging.INFO)


class LLMTransformer(TransformerMixin, Generic[InputType, OutputType]):
    """
    A scikit-learn based Transformer that uses an LLM to transform the data.

    Currently it only supports OpenAI models.
    """

    def __init__(
        self,
        input_type: Type[InputType],
        output_type: Type[OutputType],
        system_prompt: str = "",
        user_prompt_template: str = "",
        model: str = "gpt-4o-mini",
        max_concurrent_requests: int = 10,
        max_retries: int = 3,
    ):
        """
        Initialize the transformer.

        Args:
            input_type: The type of the input data (should be derived from pydantic.BaseModel).
            output_type: The type of the output data (should be derived from pydantic.BaseModel).
            system_prompt: The system prompt string (not a template).
            user_prompt_template: The user prompt template.
            model: The OpenAI GPT model to use (e.g. `gpt-4o-mini`).
            max_concurrent_requests: The maximum number of concurrent requests to avoid hitting rate limits.
            max_retries: The maximum number of retries.
        """
        self.system_prompt = system_prompt
        self.user_prompt = Template(user_prompt_template)
        self._input_type = input_type
        self._output_type = output_type
        self._llm_client = OpenAIClient(model, max_concurrent_requests, max_retries)

    def fit(self, X: any, y=None):
        return self

    async def transform_async(self, X: list[InputType]) -> list[OutputType]:
        user_prompts = [self.user_prompt.render(**x.model_dump()) for x in X]
        tasks = [
            self._llm_client.call(
                system_prompt=self.system_prompt, user_prompt=user_prompt, output_model=self._output_type
            )
            for user_prompt in user_prompts
        ]
        result = await asyncio.gather(*tasks, return_exceptions=True)

        return result

    def transform(self, X: list[InputType]) -> list[OutputType]:
        """
        Transforms input using LLM API and returns the list of output instances.
        """
        return asyncio.run(self.transform_async(X))



import asyncio

from dotenv import load_dotenv
from openai import AsyncOpenAI
from pydantic import BaseModel

load_dotenv()


class OpenAIClient:
    """
    A thin wrapper around the OpenAI API client that handles retries and
    multiple concurrent requests with a maximum limit using semaphore.
    """

    def __init__(
        self,
        model: str = "gpt-4o-mini",
        max_concurrent_requests: int = 10,
        max_retries: int = 3,
    ):
        self.model = model
        self.max_retries = max_retries
        self.max_concurrent_requests = max_concurrent_requests
        self._semaphore = asyncio.Semaphore(max_concurrent_requests)
        self._client = AsyncOpenAI()

    async def call(self, system_prompt: str, user_prompt: str, output_model: BaseModel) -> dict:
        if user_prompt == "" and system_prompt == "":
            raise ValueError("At least one of user_prompt or system_prompt must be provided.")

        for attempt in range(self.max_retries):
            async with self._semaphore:
                completion = await self._client.beta.chat.completions.parse(
                    model=self.model,
                    messages=[
                        {"role": "system", "content": system_prompt},
                        {"role": "user", "content": user_prompt},
                    ],
                    response_format=output_model,
                )
                result = completion.choices[0].message
                if result.parsed:
                    return result.parsed
                else:
                    print("Failure: ", completion)
                    await asyncio.sleep(2**attempt)

        raise Exception("Max retries exceeded")



from typing import TypeVar

from pydantic import BaseModel


class InputModel(BaseModel):

    @classmethod
    def prompt_fields(cls) -> list[str]:
        """Return the fields that should be included in the prompt.

        By default, all fields will be included. Subclass can override this to include only specific fields.
        """
        return list(cls.model_fields.keys())


InputType = TypeVar("InputModel", bound=BaseModel)
OutputType = TypeVar("OutputModel", bound=BaseModel)
